@@ -1,6 +1,7 @@
 from configparser import RawConfigParser
 from collections import defaultdict
 from functools import lru_cache
+import json
 import imp
 from logging import info, error, debug
 import os
@@ -10,9 +11,11 @@ import re
 import shutil
 from subprocess import check_output as run, CalledProcessError, STDOUT
 import sys
+from time import time, sleep
 import urllib.request
 import urllib.error
 
+from autopip.constants import UpdateFreq
 from autopip import crontab, exceptions
 
 
@@ -32,47 +35,77 @@ class AppsManager:
         # PyPI auth. Tuple of user and password.
         self._index_auth = None
 
-    def install(self, apps):
+    def install(self, apps, update=None, wait=False):
         """
         Install the given apps
 
         :param list[str] apps: List of apps to install
+        :param UpdateFreq|None update: How often to update
+        :param bool wait: Wait for a new version to be published and then install it.
         """
         self._set_index()
 
         autopip_path = shutil.which('autopip')
         if (self.paths.is_user and sys.stdout.isatty() and not list(self.apps) and autopip_path and
                 autopip_path.startswith(str(self.paths.SYSTEM_SYMLINK_ROOT))):
-            info('! Based on permission, this will install to your user home instead of %s',
+            info('# Based on permission, this will install to your user home instead of %s',
                  self.paths.SYSTEM_SYMLINK_ROOT)
             info('  To install for everyone, cancel using CTRL+C and then re-run using sudo.')
-            info('  As using sudo to install is a security risk, please do so only if you trust the app.')
 
         failed_apps = []
+        printed_wait = False
 
         for name in apps:
             try:
-                app_spec = next(iter(pkg_resources.parse_requirements(name)))
-                app, updated = self._install_app(app_spec)
+                if isinstance(name, tuple):  # From app.group_specs() or self.update()
+                    name, update = name
+                    if update:
+                        update = UpdateFreq.from_name(update)
 
-                group_specs = app.group_specs()
-                if updated and group_specs:
-                    info('This app has defined "autopip" entry points to install: %s', ' '.join(group_specs))
-                    apps.extend(group_specs)
+                app_spec = next(iter(pkg_resources.parse_requirements(name)))
+                app, updated = self._install_app(app_spec, update=update, wait=wait)
+
+                if updated:
+                    printed_wait = False
+                    group_specs = app.group_specs()
+                    if group_specs:
+                        info('This app has defined "autopip" entry points to install: %s', ' '.join(
+                             s[0] for s in group_specs))
+                        apps.extend([s for s in group_specs if s not in apps])
+
+                elif wait:
+                    goback = '\033[1A' if printed_wait else ''
+                    print(f'{goback}Waiting for new version of {name} to be published...'.ljust(80))
+                    sleep(60)
+                    apps.append(name)
+                    printed_wait = True
 
             except Exception as e:
                 error(f'! {e}', exc_info=self.debug)
                 failed_apps.append(name)
+                printed_wait = False
 
         if failed_apps:
             raise exceptions.FailedAction()
 
-    def _install_app(self, app_spec):
+    def _install_app(self, app_spec, update=None, wait=False):
         """ Install the given app """
-        version = self._app_version(app_spec)
+        app = App(app_spec.name, self.paths, debug=self.debug)
+        updated = False
 
-        app = App(app_spec.name, self.paths)
-        updated = app.install(version, app_spec)
+        # Skip update if install was done within the update frequency when run from cron
+        if (sys.stdout.isatty() or not app.is_installed or
+                update and app.path.stat().st_mtime + update.seconds < time()):
+            if app.is_installed:
+                app.path.touch()
+
+            version = self._app_version(app_spec)
+
+            if not wait or version != app.current_version:
+                updated = app.install(version, app_spec, update=update)
+
+        else:
+            debug('App does not need to be updated yet.')
 
         return app, updated
 
@@ -168,7 +201,13 @@ class AppsManager:
                 continue
 
             app_path = str(app.current_path.resolve())
-            app_info.append((app.name, app.current_version, app_path))
+
+            if app.settings().get('update'):
+                update = f"[updates {app.settings()['update']}]"
+            else:
+                update = ''
+
+            app_info.append((app.name, app.current_version, app_path, update))
 
             if scripts:
                 hide_path = False
@@ -179,7 +218,7 @@ class AppsManager:
                         if hide_path:
                             script_path = script_path.replace(str(script_symlink.parent) + '/',
                                                               ' ' * len(str(script_symlink.parent)) + ' ')
-                        app_info.append(('', '', script_path))
+                        app_info.append(('', '', script_path, ''))
                         hide_path = True
 
         if app_info:
@@ -189,7 +228,7 @@ class AppsManager:
                     info_lens[i] = len(value) if len(value) > info_lens[i] else info_lens[i]
 
             # Print table
-            table_style = '  '.join('{{:{}}}'.format(l) for l in info_lens.values())
+            table_style = '  '.join('{{:{}}}'.format(l) if l else '{}' for l in info_lens.values())
             for info_part in app_info:
                 info(table_style.format(*info_part))
 
@@ -211,7 +250,12 @@ class AppsManager:
                 continue
 
             if name == 'autopip' and len(list(self.apps)) > 1:
-                error('! autopip can not be uninstalled until other apps are uninstalled')
+                if apps[-1] == 'autopip':
+                    error('! autopip can not be uninstalled until other apps are uninstalled: %s', ' '.join(
+                        a.name for a in self.apps if a.name != 'autopip'))
+                else:  # Try again after uninstall the other apps
+                    apps.append('autopip')
+
                 continue
 
             app = App(name, self.paths)
@@ -226,6 +270,46 @@ class AppsManager:
             else:
                 info(f'{name} is not installed')
 
+        if not list(self.apps):
+            try:
+                crontab.remove('autopip')
+            except Exception as e:
+                debug(e)
+
+    def update(self, apps=None, wait=False):
+        """
+        Update installed apps
+
+        :param list apps: List of apps to update. Defaults to all.
+        :param bool wait: Wait for a new version to be published and then install it.
+        """
+        app_instances = list([a for a in self.apps if a.name in apps] if apps else self.apps)
+
+        if app_instances:
+            app_specs = []
+            for app in app_instances:
+                settings = app.settings()
+                if settings.get('update'):
+                    app_specs.append((settings['app_spec'], settings['update']))
+                elif sys.stdout.isatty():
+                    app_specs.append((settings.get('app_spec', app.name), None))
+
+            if app_specs:
+                self.install(app_specs, wait=wait)
+
+            elif not apps:
+                try:
+                    crontab.remove('autopip')
+                except Exception as e:
+                    debug(e)
+
+        elif list(self.apps):
+            info('No apps found matching: %s', ', '.join(apps))
+            info('Available apps: %s', ', '.join([a.name for a in self.apps]))
+
+        else:
+            info('No apps installed yet.')
+
 
 class App:
     """ Represents an app that may or may not be installed on disk """
@@ -233,13 +317,15 @@ class App:
     #: Prefixes of scripts to skip when creating symlinks
     SKIP_SCRIPT_PREFIXES = {'activate', 'pip', 'easy_install', 'python', 'wheel'}
 
-    def __init__(self, name, paths):
+    def __init__(self, name, paths, debug=False):
         """
         :param str name: Name of the app
         :param AppsPath paths: Path paths
+        :param bool debug: Turn on debug mode
         """
         self.name = name
         self.paths = paths
+        self.debug = debug
 
         #: Path to install all app versions
         self.path = self.paths.install_root / name
@@ -266,32 +352,32 @@ class App:
         if self.current_path:
             return self.current_path.resolve().name
 
-    def install(self, version, app_spec):
+    def install(self, version, app_spec, update=None):
         """
         Install the version of the app if it is not already installed
 
         :param str version: Version of the app to install
         :param pkg_resources.Requirement app_spec: App version requirement from user
+        :param UpdateFreq|None update: How often to update. Choose from hourly, daily, weekly, monthly
         :return: True if install or update happened, otherwise False when nothing happened (already installed / non-tty)
         """
         version_path = self.path / version
         prev_version_path = self.current_path and self.current_path.resolve()
         important_paths = [version_path, prev_version_path, self._current_symlink]
 
-        if not shutil.which('curl'):
-            raise exceptions.MissingCommandError('curl is not available and is required to install pip. '
-                                                 'Please install and then re-run')
-
         if version_path.exists():
             if self.current_version == version:
                 # Skip printing / ensuring symlinks / cronjob when running from cron
                 if not sys.stdout.isatty():
                     return False
 
-                info(f'{self.name} is already installed')
+                pinned = str(app_spec).lstrip(self.name)
+                pin_info = f' [per spec: {pinned}]' if pinned else ''
+
+                info(f'{self.name} is up-to-date{pin_info}')
 
             else:
-                info(f'Setting {version} as the current version for {self.name}')
+                info(f'{self.name} {version} was previously installed and will be set as the current version')
 
         else:
             old_venv_dir = None
@@ -306,9 +392,9 @@ class App:
                                                           if os.path.exists(p) and not p.startswith(old_venv_dir)])
 
                 run(f"""set -e
-                    python3 -m venv {version_path} --without-pip
+                    python3.6 -m venv {version_path}
                     source {version_path / 'bin/activate'}
-                    curl -s https://bootstrap.pypa.io/get-pip.py | python
+                    pip install --upgrade pip
                     pip install {self.name}=={version}
                     """, executable='/bin/bash', stderr=STDOUT, shell=True)
 
@@ -345,19 +431,54 @@ class App:
                 '  If you are the app owner, make sure to setup entry_points in setup.py.\n'
                 '  See http://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation')
 
-        # Install cronjobs
-        if sys.stdout.isatty():  # Skip updating cronjob when run from cron
-            try:
-                autopip_path = shutil.which('autopip')
-                if not autopip_path:
-                    raise exceptions.MissingCommandError(
-                        'autopip is not available. Please make sure its bin folder is in PATH env var')
-                crontab.add(f'{autopip_path} install "{app_spec}" 2>&1 >> {self.paths.log_root / "cron.log"}',
-                            cmd_id=self._crontab_id)
-                info('Auto-update enabled via cron service')
+        self.settings(app_spec=str(app_spec))
 
-            except Exception as e:
-                error('! Auto-update was not enabled because: %s', e)
+        # Install cronjobs
+        if 'update' not in sys.argv:
+            pinning = '==' in str(app_spec) and not str(app_spec).endswith('*')
+            if pinning and (self.settings().get('update') or update):
+                info('Auto-update will be disabled since we are pinning to a specific version.')
+                info('To enable, re-run without pinning to specific version with --update option')
+
+                if self.settings().get('update'):
+                    self.settings(update=None)
+                    try:
+                        crontab.remove(self._crontab_id)
+                    except exceptions.MissingError as e:
+                        debug(e)
+
+            elif update:
+                try:
+                    autopip_path = shutil.which('autopip')
+                    if not autopip_path:
+                        raise exceptions.MissingError(
+                            'autopip is not available. Please make sure its bin folder is in PATH env var')
+
+                    # Migrate old crontabs
+                    try:
+                        old_crons = [c for c in crontab.list().decode().split('\n') if c and 'autopip update' not in c]
+                        if old_crons:
+                            cron_re = re.compile('autopip install "(.+)"')
+                            for cron in old_crons:
+                                match = cron_re.search(cron)
+                                if match:
+                                    old_app_spec = next(iter(pkg_resources.parse_requirements(match.group(1))))
+                                    old_app = App(old_app_spec.name, self.paths)
+                                    if old_app.is_installed:
+                                        old_app.settings(app_spec=str(old_app_spec))
+                            crontab.remove('autopip')
+
+                    except Exception as e:
+                        debug(e)
+
+                    crontab.add(f'{autopip_path} update '
+                                f'2>&1 >> {self.paths.log_root / "cron.log"}', cmd_id='autopip update')
+                    info(update.name.title() + ' auto-update enabled via cron service')
+
+                    self.settings(update=update.name.lower())
+
+                except Exception as e:
+                    error('! Auto-update was not enabled because: %s', e, exc_info=self.debug)
 
         # Install script symlinks
         prev_scripts = self.scripts(prev_version_path) if prev_version_path else set()
@@ -396,11 +517,30 @@ class App:
                 script_symlink.unlink()
                 info('- '.format(script_symlink.name))
 
-        if not printed_updating and sys.stdout.isatty() and current_scripts:
+        if not printed_updating and sys.stdout.isatty() and current_scripts and 'update' not in sys.argv:
             info('Scripts are in {}: {}'.format(self.paths.symlink_root, ', '.join(sorted(current_scripts))))
 
         return True
 
+    def settings(self, **new_settings):
+        """ Get or set settings """
+        current_settings = {}
+
+        if self.path.exists():
+            settings_file = self.path / 'settings.json'
+            if settings_file.exists():
+                try:
+                    current_settings.update(json.load(settings_file.open()))
+                except Exception as e:
+                    debug('Could not load app settings: %s', e)
+
+            if new_settings:
+                current_settings.update(new_settings)
+                with settings_file.open('w') as fh:
+                    json.dump(current_settings, fh)
+
+        return current_settings
+
     def scripts(self, path=None):
         """ Set of scripts for the given app path (defaults to current). """
         dist = self._pkg_distribution(path=path)
@@ -437,10 +577,18 @@ class App:
 
         if dist:
             for app, spec in dist.get_entry_map('autopip').items():
-                if spec.module_name == 'latest' or name_only:
+                if name_only:
                     app_specs.append(app)
+
+                elif spec.module_name == 'latest':
+                    app_specs.append((app, UpdateFreq.DEFAULT.name.lower()))
+
                 else:
-                    app_specs.append(f'{app}=={spec.module_name}.*')
+                    if len(spec.module_name.split('.')) < 3 or spec.extras:     # Wildcard with auto-update
+                        app_specs.append((f'{app}=={spec.module_name}.*',
+                                         next(iter(spec.extras), UpdateFreq.DEFAULT.name.lower())))
+                    else:
+                        app_specs.append((f'{app}=={spec.module_name}', None))  # Specific version without auto-update
 
         return app_specs
 
@@ -489,7 +637,10 @@ class App:
         """ Uninstall app """
         info('Uninstalling %s', self.name)
 
-        crontab.remove(self._crontab_id)
+        try:
+            crontab.remove(self._crontab_id)
+        except exceptions.MissingError as e:
+            debug(e)
 
         for script in self.scripts():
             script_symlink = self.paths.symlink_root / script
@@ -557,13 +708,15 @@ class AppsPath:
 
         # Check local
         if system_reasons:
-            if not os.access(self.LOCAL_INSTALL_ROOT.parent, os.W_OK):
+            if not (os.access(self.LOCAL_INSTALL_ROOT.parent, os.W_OK) or
+                    not self.LOCAL_INSTALL_ROOT.parent.exists() and
+                    os.access(self.LOCAL_INSTALL_ROOT.parent.parent, os.W_OK)):
                 local_reasons.append(f'No permission to write to {self.LOCAL_INSTALL_ROOT.parent}')
 
             if not os.access(self.LOCAL_SYMLINK_ROOT, os.W_OK):
                 local_reasons.append(f'No permission to write to {self.LOCAL_SYMLINK_ROOT}')
 
-            if not (self.LOCAL_LOG_ROOT.parent.exists() and os.access(self.LOCAL_LOG_ROOT.parent, os.W_OK) or
+            if not (os.access(self.LOCAL_LOG_ROOT.parent, os.W_OK) or
                     not self.LOCAL_LOG_ROOT.parent.exists() and os.access(self.LOCAL_LOG_ROOT.parent.parent, os.W_OK)):
                 local_reasons.append(f'No permission to write to {self.LOCAL_LOG_ROOT.parent}')
 
